#!/usr/bin/env python
#title:		cnn_processing.py
#description:	The script performs the CNN classification with a specified model and stores the top 3 classification results in a proper Sqlite database.
#author:	A. Ortis
#date:		20170925
#version:		0.1
#usage:		call the function	cnn_process_directory(CNN_MODEL, DIR, DBname)
#
#			Parameters:
#				- CNN_MODEL:	specify a model among AlexNet, GoogleNet, MVSOenglish or Places205
#				- DIR:		directory containing the image dataset to process
#				- DBname:		database used to store the results

#notes:		the script needs a configuration.ini with the caffe path and the paths of the pre-trained CNN models. An example of configuration.ini file is provided.
#==============================================================================


import warnings

warnings.filterwarnings("ignore")

import os
import string
import numpy as np
import sys
os.environ["minloglevel"] = "3"
os.environ["GLOG_minloglevel"] = "3"
import caffe
import glob
import time
import json
from ConfigParser import SafeConfigParser
import sqlite3 as lite


from contextlib import contextmanager
import io, sys


caffe_root = ''
labels_source = ''
delim = ''
cnn_name = ''
transformer = ''

def cnn_process_directory(CNN_MODEL, DIR, DBname):


	#Image list, all the images are stored in a given directory
	img_list = os.listdir(DIR)

	convnet = cnn_network_init(CNN_MODEL)
	for idx,picID in enumerate(img_list):
		picPath = DIR+"/"+picID
		print "Processing image\t"+ picID
		if not os.path.isfile(picPath):
			print "image "+picPath+ " doesn't exist!"
			sys.exit(1)
		
		with warnings.catch_warnings():
			warnings.simplefilter("ignore")
			top_k_classification, top_k_labels, probs = cnn_extract_features(convnet, picPath)

		print "DB connection..."
		con = lite.connect(DBname)
		with con:
			cur = con.cursor()
			cur.execute("INSERT INTO Cnndata(FlickrId, Cnn, top_3_labels, top_3_probs) VALUES (?, ?, ?, ?)",(picID, CNN_MODEL,  json.dumps(top_k_labels.tolist()),json.dumps(probs.tolist())))
		con.close()
		print "Last img processed has idx:\t" + str(idx)
		print "DB connection closed.\n\n"



#	return feat, top_k_classification, top_k_labels, probs 

def cnn_network_init(CNN_MODEL):

	parser = SafeConfigParser()
	#config_path = os.path.realpath('configuration.ini')
	config_path = os.path.dirname(os.path.realpath(sys.argv[0]))
	config_path = config_path + '/configuration.ini'
	parser.readfp(open(config_path))
	
	# Read configuration parameters
	global caffe_root
	caffe_root = parser.get('general','caffe_root')
	global labels_source
	labels_source = parser.get(CNN_MODEL,'labels_source')	
	global delim	
	delim = parser.get(CNN_MODEL,'delim')	
	global cnn_name	
	cnn_name = parser.get(CNN_MODEL,'cnn_name')
	prototxt = parser.get(CNN_MODEL,'prototxt')
	cnn_model = parser.get(CNN_MODEL,'cnn_model')
	mean_image = parser.get(CNN_MODEL,'mean_image')
	input_w = int(parser.get(CNN_MODEL,'input_w'))	
	input_h = int(parser.get(CNN_MODEL,'input_h'))	
	
	# Make sure that caffe is on the python path:
	sys.path.insert(0, caffe_root + 'python')
	
	# CNN configuration
	caffe.set_mode_cpu()
	#global net
	
	net = caffe.Net(caffe_root + prototxt,
			caffe_root + cnn_model,
			caffe.TEST)
	

	# Input preprocessing: 'data' is the name of the input blob == net.inputs[0]
	global transformer
	transformer = caffe.io.Transformer({'data': net.blobs['data'].data.shape})
	transformer.set_transpose('data', (2,0,1))

	if cnn_name == 'AlexNet' or cnn_name =='MVSOenglish' or cnn_name =='GoogleNet':
		transformer.set_mean('data', np.load(caffe_root + mean_image).mean(1).mean(1)) # mean pixel
	elif cnn_name == 'Places205':
		mean_image = json.loads(mean_image)  #from string to list
		transformer.set_mean('data', np.array(mean_image)) # mean pixel

	transformer.set_raw_scale('data', 255)  		# the reference model operates on images in [0,255] range instead of [0,1]
	transformer.set_channel_swap('data', (2,1,0)) 	# the reference model has channels in BGR order instead of RGB

	net.blobs['data'].reshape(1,3,input_h,input_w)

	return net


def cnn_extract_features(net, INPUT_FILE):

	net.blobs['data'].data[...] = transformer.preprocess('data', caffe.io.load_image(INPUT_FILE))
	out = net.forward()
	classification = format(out['prob'].argmax())
	#print("Predicted class is #{}.".format(out['prob'].argmax()))
	#print("Predicted class is #."+classification)


	labels_filename = caffe_root + labels_source
	
	labels = np.loadtxt(labels_filename, str, delimiter=str(delim))
	
	
	# sort top k predictions from softmax output
	top_k = net.blobs['prob'].data[0].flatten().argsort()[-1:-4:-1] # takes the first three

	# The indices start from zero
	print labels[top_k]
	probs = net.blobs['prob'].data[0].flatten()[top_k]

	return top_k, labels[top_k], probs
	





